--[[
	WindyLoads.lua
	
	Author: 		Sushifresser
	Date: 		06.02.2022
	Version: 	0.0.3.0
	
	Changelog:	V 0.0.3.0 @ 06.02.2022 - first implementation of mod configuration via XML modDesc. Preparation for first publishing
	Changelog:	V 0.0.2.0 @ 10.12.2021 - wind is now included in the calculation of the air speed above the vehicle
	Changelog:	V 0.0.1.0 @ 09.12.2021 - initial implementation in FS 22
]]

--[[Plan fürs Programmieren des Mods
- bei initialisierung prüfen, ob das aktuelle vehicle überhaupt geeignet ist um den Mod anzuwenden
	- sollte eine fillUnit haben
	- sollte anwendbaren FillType haben
	- sollte nicht für Flüssigkeiten gedacht sein
	- das Ergebnis der Prüfung in einer eigenen Variable speichern um bei Updatefunktion schnell entscheiden zu können, ob gerechnet werden muss
	- Die Liste der erlaubten/anwendbaren FillTypes könnte per Definition in der XML-Datei erweitert werden
- Ermitteln der aktuellen Geschwindigkeit des Spielers
	- lässt sich evtl. auch die Geschwindigkeit eines Fahrzeugs ermitteln?
- Prüfen ob das Fahrzeug den korrekten Typ hat (Trailer, mit offener Plane)
- Ermitteln welchen Fülltyp das Fahrzeug hat
- Aus Fülltyp und Geschwindigkeit eine Rate des wegblasens ermitteln
	- später könnte man noch die Windgeschwindigkeit mit eingeziehen
	- der relative Füllstand des Fahrzeugs könnte auch einen Einfluss auf dei Rate haben. Volle Anhänger werden sicherlich mehr verlieren als fats leere Anhänger.
- Die Rate des Wegblasens auf den Füllinhalt des Fahrzeugs anwenden
	- dieses Event an alle Spieler im Multiplayer verteilen
	- der Inhalt soll nicht auf dem Boden landen, sondern einfach verschwinden
]]


--[[Sammlung von Informationen zum Programmieren
- Abfrage der Objektgeschwindigkeit: vx, vy, vz = getLinearVelocity(object)
- setzen der linearen Geschwindigkeit: setLinearVelocity(object, vx, vy, vz)
- Geschwindigkeitsangaben werden in der Engine in m/s gehandhabt
- Geschwindigkeit eines Vehicle lässt sich durch "getLastSpeed()" ermitteln. Der Rückgabewert ist ein float in km/h
]]

--[[TODOs
- wenn man feststellt, dass ein Fahrzeug nicht weiter betrachtet werden muss für diesen Mod, könnte man es auch den Events usw. abmelden um nicht weiter Rechenzeit auf dieses FGahrzeug zu vergeuden.
--später ggf. auch noch erlauben Fahrzeuge manuell per XML config hinzuzufügen.
-- man könnte die Anhängerneigung noch mit einbeziehen beim herauswehen
-- man könnte auch simulieren, dass umgekippte Anhänger sich entleeren
-- Grenzwert ab wann etwas herausgeblasen wird, könnte auch per config einstellbar sein
-- über config könnte globaler Faktor einstellbar sein, welcher den Effekt des herausblasens für alles skaliert

]]

WindyLoads = {};
local modName = g_currentModName;
local modDirectory = g_currentModDirectory;

source(modDirectory .. "WindyLoadsBlowEvent.lua");

WindyLoads.MIN_SPEED_TO_BLOW = 20;	--the minimum speed below which no loads will be blown away
WindyLoads.DEBUG_MODE = false;	--determines whether debug information will be provided
WindyLoads.CONSIDER_WIND = true;	--determines whether the wind speed will be taken into account when calculating the relative air speed above the vehicle.

WindyLoads.modConfigDone = false;	--marker for the completed initial configuration of this mod

WindyLoads.SYNC_TIME_INTERVAL_MS = 1000;	--the amount of milliseconds that is to pass between two synchronisation events of the fill unit levels

WindyLoads.currentModDirectory = g_currentModDirectory;
WindyLoads.modDesc = loadXMLFile("modDesc", WindyLoads.currentModDirectory .. "modDesc.xml");

WindyLoads.hudData = {
	hudTextPosX = 0.99,
	hudTextPosY = 0.85,
	hudTextSize = 0.015 * g_gameSettings.uiScale
};
WindyLoads.applicableFillTypeCategories = {
--all FillTypeCategories, that are considered for this mod
	"LOADINGVEHICLE",
	"HAYLOFT",
	"WINDROW",
	"COMBINE",
	"FORK"
};

WindyLoads.excludedFillTypeCategories = {
--all FillTypeCategories which are to be excluded from this mod, even if they are defined in the applicable FillTypeCategories
	"LIQUID",
	"SPRAYER",
	"ANIMAL",
	"HORSE",
	"PRODUCT",
	"SLURRYTANK",
	"PIECE"
};

-- the standard weight per liter for unknown fill types
WindyLoads.STANDARD_MASS_PER_LITER = 0.0006;

-- the standard blow factor for unknown fill types
WindyLoads.STANDARD_BLOW_FACTOR = 0.5; 

WindyLoads.blowFactors = {
--this table stores factors for each FillType that determines how strong the effect of the wind is.
--lower factors make the load to be blown away slower
	UNKNOWN = WindyLoads.STANDARD_BLOW_FACTOR,	--this is just to be sure
	WHEAT = 0.5,
	BARLEY = 0.5,
	OAT = 0.55,
	CANOLA = 0.5,
	SORGHUM = 0.5,
	SUNFLOWER = 0.6,
	SOYBEAN = 0.4,
	MAIZE = 0.3,
	SEEDS = 0.7,
	FORAGE = 0.5,
	CHAFF = 0.7,
	WOODCHIPS = 0.4,
	SILAGE = 0.15,
	GRASS_WINDROW = 0.8,
	DRYGRASS_WINDROW = 1,
	STRAW = 1,
	FERTILIZER = 0.15,
	MANURE = 0.05,
	LIME = 0.2
};


--this is to store the relation between index and name of a FillType. Can be omitted later to work with the values given by the managers of the game.
--key: (int) fill type index; value: (string) fill type name
WindyLoads.fillTypesIndexToName = {};

--this will store all FillTypes which are considered applicable for this mod
--key: (string) fill type name; value: (table) fillType
WindyLoads.applicableFillTypes = {}; 

function WindyLoads.prerequisitesPresent(specializations)
	local applicable = false;
	applicable = SpecializationUtil.hasSpecialization(FillUnit, specializations) and SpecializationUtil.hasSpecialization(FillVolume, specializations) and SpecializationUtil.hasSpecialization(Wheels, specializations);
	applicable = applicable and SpecializationUtil.hasSpecialization(Dischargeable, specializations);
	return applicable;
end;

function WindyLoads.registerEventListeners(vehicleType)
	local functionNames = {
		"onLoad",
		"onUpdate",	
		"onDraw", --only for debug hud so far
		"onRegisterActionEvents"
	};
	for _, functionName in ipairs(functionNames) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, WindyLoads);
		--print(string.format("registered eventlistener %s for vehicle Type %s", functionName, vehicleType.name));
	end;
end;

function WindyLoads:onRegisterActionEvents(isSelected, isOnActiveVehicle)
    local actionEventId;
    if isOnActiveVehicle then
        _, actionEventId = InputBinding.registerActionEvent(g_inputBinding, InputAction.WINDY_DEBUG, self, WindyLoads.actionEventWindyDebug, false, true, false, true, nil, true);
		--g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL);
		self.spec_windyLoads.actionEventDebugId = actionEventId;
		--print(string.format("[WindyLoads]: successfully registered action event with ID: %s", actionEventId));
    end
end

function WindyLoads.registerFunctions(vehicleType)
	--SpecializationUtil.registerFunction(vehicleType, "getVehicleIsSupportedByWindyLoads", WindyLoads.getVehicleIsSupportedByWindyLoads);
	SpecializationUtil.registerFunction(vehicleType, "getVehicleIsSupportedByWindyLoads", WindyLoads.getVehicleIsSupportedByWindyLoads);
	SpecializationUtil.registerFunction(vehicleType, "blowFromFillUnit", WindyLoads.blowFromFillUnit);
end;

function WindyLoads:actionEventWindyDebug(actionName, inputValue, callbackState, isAnalog)
	if not WindyLoads.DEBUG_MODE then
		return;
	end;
	print("[WindyLoads] : Received ActionEventWindyDebug");
	--local vehicle = self.selectionObject.vehicle;
	--self:emptyBaler(vehicle, false);
	--WindyLoads:print_table(g_currentMission.controlledVehicle, "g_currentMission.controlledVehicle");
	--WindyLoads:print_table(self.spec_windyLoads, "self.spec_windyLoads");
	--WindyLoads:print_table({WindyLoads.hudData}, "WindyLoads.hudData");

end

function WindyLoads:onLoad(savegame)
	--local specCover = self.spec_cover;
	self.spec_windyLoads = {};
	local spec = self.spec_windyLoads;
	--this stores the last calculated vehicle speed
	spec.lastVehicleSpeed = 0;

	--this table stores all fillUnits of the vehicle, that are susceptible to the mod's effects.
	--key: fillUnitIndex; value: table fillUnit
	spec.supportedFillUnits = {};
	
	--this table stores all blow rates of the corresponding fillUnits
	--key: fillUnitIndex; value: float blowRate 
	spec.blowRatePerFillUnit = {};
	
	--this table holds the last blown volume values per fill unit
	--key: fillUnitIndex; value: float volume blown away in last update tick 
	spec.lastBlownVolume = {};
	
	--this variable stores the passed time since the last synchronisation of fill levels for each fill unit indiviually
	spec.timeSinceLastSync = {};
	
	--just to be sure, we save the reference to our vehicle here
	spec.vehicle = self;
	
	if not WindyLoads.modConfigDone then 
		--try to find all fillTypes in the game that this mod can be applied to. They will be stored in the mod's class.
		WindyLoads.findApplicableFillTypes();
		--try to read the XML config to configure this mod with user settings
		WindyLoads:getXmlConfig() 
		--now we're done
		WindyLoads.modConfigDone = true;
		delete(WindyLoads.modDesc);
	end;
	
	--a table to hold several debug information
	spec.debugInfo = {};
	
	spec.modInitialized = true;
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Mod successfully loaded for vehicle '%s.'", Vehicle.getFullName(self)));
	end;
	
end;

--this function tries to extract some configuration values from the mod description XML file.
function WindyLoads:getXmlConfig()

	local modDesc = WindyLoads.modDesc;
	
	WindyLoads.DEBUG_MODE = Utils.getNoNil(getXMLBool(modDesc, "modDesc.windyLoadsConfigurations#debugMode"), WindyLoads.DEBUG_MODE);
	WindyLoads.CONSIDER_WIND = Utils.getNoNil(getXMLBool(modDesc, "modDesc.windyLoadsConfigurations#considerWindSpeed"), WindyLoads.CONSIDER_WIND);
	local specializationNumber = 0;
	while true do
		local specializationKey = string.format("modDesc.windyLoadsConfigurations.fillTypeConfigurations.fillTypeConfiguration(%d)", specializationNumber);
		if not hasXMLProperty(modDesc, specializationKey) then
			break;
		end;
		local fillTypeName = (Utils.getNoNil(getXMLString(modDesc, specializationKey .. "#fillTypeName"), "")):upper();
		if fillTypeName ~= "" then
			local fillTypeInclude = Utils.getNoNil(getXMLBool(modDesc, specializationKey .. "#fillTypeInclude"), false);
			local fillTypeExclude = Utils.getNoNil(getXMLBool(modDesc, specializationKey .. "#fillTypeExclude"), false);
			local fillTypeBlowFactor = Utils.getNoNil(getXMLFloat(modDesc, specializationKey .. "#fillTypeBlowFactor"), WindyLoads.STANDARD_BLOW_FACTOR);
			if fillTypeInclude and fillTypeExclude then
			--only one of the attributes should be provided. We will handle this item as to be excluded.
				Logging.xmlWarning(modDesc, "[WindyLoads]: WindyLoads recieved same fillType '%s' to be included an excluded at the same time. Item will be handled as to be excluded. Please check your configuration!", fillTypeName);
				fillTypeInclude = false;
			end;
			if fillTypeExclude then
				if WindyLoads.applicableFillTypes[fillTypeName] ~= nil then
					WindyLoads.fillTypesIndexToName[WindyLoads.applicableFillTypes[fillTypeName].index] = nil;
					WindyLoads.applicableFillTypes[fillTypeName] = nil;
					WindyLoads.blowFactors[fillTypeName] = nil;
					if WindyLoads.DEBUG_MODE then
						print(string.format("[WindyLoads]: Removed FillType '%s' from applicable FillTypes of WindyLoads mod based on user-defined XML configuration.", fillTypeName));
					end;
				end
			elseif fillTypeInclude then
				if WindyLoads.applicableFillTypes[fillTypeName] == nil then
				-- the fill type needs to be added at first
					local fillType = g_fillTypeManager:getFillTypeByName(fillTypeName);
					if fillType~= nil and fillType.name ~= nil then
						WindyLoads.applicableFillTypes[fillType.name] = fillType;
						WindyLoads.fillTypesIndexToName[fillType.index] = fillType.name;
						if WindyLoads.DEBUG_MODE then
							print(string.format("[WindyLoads]: Added FillType '%s' to applicable FillTypes of WindyLoads mod based on user-defined XML configuration.", fillType.name));
						end;
					end;
				end;
				if WindyLoads.applicableFillTypes[fillTypeName] == nil then
				-- we know the fill type already and now we can adjust its parameters
					WindyLoads.blowFactors[fillTypeName] = fillTypeBlowFactor;
				end;
			end;
		end;
			
		specializationNumber = specializationNumber + 1;
	end;
end;


--to get the velocity of the current vehicle
function WindyLoads:getVehicleVelocity()
	local spec = self.spec_windyLoads;
	if not spec.modInitialized then
		return;
	end;
	local velocity = self:getLastSpeed();
	if velocity == nil or not velocity or velocity < 0 then 
		return 0;
	end
	return velocity;
end;

function WindyLoads:findApplicableFillTypes()
--this function determines all applicable fillTypes based on the definitions of the FillTypeCategories
	local applicableFillTypes = {};
	local fillTypesToCheck = g_fillTypeManager:getFillTypes();
	Logging.devInfo("[WindyLoads]: Received " .. #fillTypesToCheck .. " from FillType Manager to check.");
	--WindyLoads:print_table(fillTypesToCheck, "fillTypesToCheck", "[WindyLoads]: The following fill types have been detected in the game: ");
	for _,fillType in pairs(fillTypesToCheck) do
		if WindyLoads.applicableFillTypes[fillType.name] == nil then
		--only go on if the FillType at hand is not already known
			local isApplicable = false;
			for _, applicableFillTypeCategory in pairs(WindyLoads.applicableFillTypeCategories) do
				if g_fillTypeManager:getIsFillTypeInCategory(fillType.index, applicableFillTypeCategory) then
					isApplicable = true;
					break;
				end;
			end;
			if isApplicable then
			--we found a FillType that seems to be applicable. Now we have to check if it is still to be excluded.
				for _, excludedFillTypeCategory in pairs(WindyLoads.excludedFillTypeCategories) do
					if g_fillTypeManager:getIsFillTypeInCategory(fillType.index, excludedFillTypeCategory) then
						isApplicable = false;
						break;
					end;
				end;
			end;
			if isApplicable then
			--the current fillType has passed all checks and can now be considered applicable
				WindyLoads.applicableFillTypes[fillType.name] = fillType;
				WindyLoads.fillTypesIndexToName[fillType.index] = fillType.name;	-- this table is for more comfortable access to the data
				if WindyLoads.DEBUG_MODE then
					print(string.format("[WindyLoads]: Added FillType '%s' to applicable FillTypes of WindyLoads mod based on standard configuration.", fillType.name));
				end;
			end;
		end;
	end;
	--WindyLoads:print_table(WindyLoads.applicableFillTypes, "WindyLoads.applicableFillTypes", "[WindyLoads]: The following fill types have been registered to be applicable for this mod: ");
end;

---Checks if the vehicle is valid for this mod to be applied to
-- @param table optional vehicle: The reference to the vehicle that is to be examined
-- @return boolean true if the vehicle is supported by this mod. Returns false otherwise 
-- @return table Returns table of supported fillUnits of the vehicle if vehicle is supported. Returns nil otherwise.
function WindyLoads:getVehicleIsSupportedByWindyLoads(vehicle)
--this function should chek if this vehicle is applicable for this mod
--TODO: maybe find some more useful criteria for checking the vehicles
	-- some vehicle types could be further excluded by default, e.g. combines, balers?
	local self = vehicle or self;
	local spec = self.spec_windyLoads;
	local specFillUnit = self.spec_fillUnit;
	local vehicleSupported = false;
	if specFillUnit == nil then
		if WindyLoads.DEBUG_MODE then
			print("[WindyLoads]: Could not find any FillUnit spec on this vehicle. Vehicle will not be considered for this mod any further");
		end;
		return false;
	end;
	local fillUnits = specFillUnit:getFillUnits();
	if fillUnits == nil or  not (#fillUnits > 0) then
		if WindyLoads.DEBUG_MODE then
			print("[WindyLoads]: Could not find any FillUnits on this vehicle. Vehicle will not be considered for this mod any further");
		end;
		return false;
	end;
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Found " .. #fillUnits .. " FillUnits on vehicle '%s'.", Vehicle.getFullName(self)));	
	end;
	local supportedFillUnits = {};
	for fillUnitIndex, fillUnit in pairs(fillUnits) do
		local fillUnitSupported = false;
		local supportedFillTypes = specFillUnit:getFillUnitSupportedFillTypes(fillUnitIndex);
		if supportedFillTypes ~= nil then
			for fillTypeIndex, _ in pairs(supportedFillTypes) do
			--loop through all supported fillUnitFillTypes to find out if one is included which we can support with our mod
				for fillTypeName, fillType in pairs(WindyLoads.applicableFillTypes) do
					if fillType.index == fillTypeIndex then
						fillUnitSupported = true;
						break;
					end;
				end;
				if fillUnitSupported then
				-- we found the current fillUnit to be supported by our mod. Save it and then go on to the next fillUnit.
					supportedFillUnits[fillUnitIndex] = fillUnit;
					spec.blowRatePerFillUnit[fillUnitIndex] = 0;	--initialize the blow rates table
					spec.timeSinceLastSync[fillUnitIndex] = 0;		--initialize the sync times table
					spec.lastBlownVolume[fillUnitIndex] = 0;		--initialize the blow volumes table
					vehicleSupported = true;
					break;
				end;
				
			end;
		end;
		--if vehicleSupported then break end;
	end;
	if WindyLoads.DEBUG_MODE then
		print(string.format("[WindyLoads]: Vehicle '%s' was %s checked for being applicable for WindyLoads mod.", Vehicle.getFullName(self), (vehicleSupported and "successfully" or "not successfully")));
	end;
	--WindyLoads:print_table(supportedFillUnits, "supportedFillUnits", "The following FillUnits were found on the vehicle");
	--if self.spec_cover then	WindyLoads:print_table(self.spec_cover, "spec_cover", "The follwing cover specification was found:") end;
	return vehicleSupported, supportedFillUnits;
end;


---This function calculates the curretn relative air speed above the vehicle. This is composed of the current vehicle velocity and direction above ground as well as the curretn wind speed and direction above ground.
-- @param table vehicle: The reference to the vehicle that is to be considered. Defaults to the own vehicle of this mod.
function WindyLoads:calculateCurrentRelativeAirSpeed(vehicle)
	local self = vehicle or self;
	local spec = self.spec_windyLoads;
	local vehicleSpeed = self:getLastSpeed();
	spec.lastVehicleSpeed = vehicleSpeed or 0;
	local airSpeed = 0;
	if WindyLoads.CONSIDER_WIND then
		local windDirX, windDirZ, windVelocity = g_currentMission.environment.weather.windUpdater:getCurrentValues();
		spec.debugInfo.windDirX = windDirX or 0;
		spec.debugInfo.windDirZ = windDirZ or 0;
		spec.debugInfo.windVelocity = windVelocity or 0;
		local vehicleWorldDirection = {};
		vehicleWorldDirection.X, vehicleWorldDirection.Y, vehicleWorldDirection.Z = self:getVehicleWorldDirection();
		spec.debugInfo.vehicleWorldDirectionX = vehicleWorldDirection.X or 0;
		spec.debugInfo.vehicleWorldDirectionY = vehicleWorldDirection.Y or 0;
		spec.debugInfo.vehicleWorldDirectionZ = vehicleWorldDirection.Z or 0;
		
		--WindyLoads:print_table(spec.debugInfo, "debugInfo");
		--WindyLoads:print_table({vehicleWorldDirection}, "vehicleWorldDirection");
		--now calculate the effective relative air spedd above the vehicle
		airSpeed = math.sqrt(vehicleSpeed^2 + windVelocity^2 + 2*windVelocity*vehicleSpeed*(vehicleWorldDirection.X*-windDirX+vehicleWorldDirection.Z*-windDirZ));
	else
		airSpeed = vehicleSpeed;
	end;
	spec.lastAirSpeed = airSpeed;
	return airSpeed;
end;

--this function calculates a value which represents the rate at which the load of a vehicle fill unit is blown away by the wind.
-- @param float airSpeed: this parameter states the air speed above the fill unit that is used to compute the blow rates
-- @param table fillUnits: A table that contains the fill unit indices as keys and the fillUnits to be examined as values. Defaults to an emtpy table.
-- @return table: Returns a table of blowrates (float values) for each FillUnit of the vehicle. The value represents percent of the current fill level per second that are to be blown away.
function WindyLoads:calculateBlowRates(airSpeed, fillUnits)
--TODO: wenn das FillVolume von einer geschlossenen Abdeckung geschützt ist, dann die blowRate auf Null setzen
	local fillUnits = fillUnits or {};
	local specFillUnit = self.spec_fillUnit;
	local blowRatePerFillUnit = {};
	for fillUnitIndex, fillUnit in pairs(fillUnits) do
		local fillLevelPercent = specFillUnit:getFillUnitFillLevelPercentage(fillUnitIndex);
		local fillUnitFillType = specFillUnit:getFillUnitFillType(fillUnitIndex);
		--WindyLoads:print_table({fill = fillUnitFillType}, "fillUnitType:");
		if airSpeed < WindyLoads.MIN_SPEED_TO_BLOW or fillUnitFillType == nil then
		--nothing will be blown away below the global minimum speed
			blowRatePerFillUnit[fillUnitIndex] = 0;
		elseif WindyLoads.fillTypesIndexToName[fillUnitFillType] == nil or fillLevelPercent == 0 then
			blowRatePerFillUnit[fillUnitIndex] = 0;
		elseif self.spec_cover and not specFillUnit:getFillUnitSupportsToolType(fillUnitIndex, ToolType.TRIGGER) then
		--this happens when there is a cover that is closed. We don't want anything to be blown away then.
			blowRatePerFillUnit[fillUnitIndex] = 0;
		else
		--now we have to calculate for each fillUnit how fast the contents are blown away
			local blowRate = 0;
			local fillUnitCapacity = (specFillUnit:getFillUnitCapacity(fillUnitIndex) or 1);
			local fillUnitFillLevel = specFillUnit:getFillUnitFillLevel(fillUnitIndex) or 0;
			-- we want the blow effect to be dependent on the fill level in a cubic manner.
			local blowVolumeFactor = fillLevelPercent ^ 3;
			local blowFactor = WindyLoads.blowFactors[WindyLoads.fillTypesIndexToName[fillUnitFillType]] or WindyLoads.STANDARD_BLOW_FACTOR;
			local fillWeight = WindyLoads.applicableFillTypes[WindyLoads.fillTypesIndexToName[fillUnitFillType]].massPerLiter or WindyLoads.STANDARD_MASS_PER_LITER;
			--we want the blow rate to depend on the air speed in a quadratic manner
			blowRate = blowFactor / fillWeight /100000000 * (airSpeed - WindyLoads.MIN_SPEED_TO_BLOW)^2;
			blowRatePerFillUnit[fillUnitIndex] = blowVolumeFactor * blowRate;
		end;
	end;
	return blowRatePerFillUnit;
end;

--this function applies the effects of this mod to all applicable fillUnits of the vehicle
function WindyLoads:affectFillUnits(dt, sendNoEvent)
	local spec = self.spec_windyLoads;
	local specFillUnit = self.spec_fillUnit;
	if not spec.modInitialized or (spec.vehicle~=nil and spec.vehicleSupported == false) then
		--there's nothing here for us to compute
		return;
	end;
	local specFillUnit = self.spec_fillUnit;
	local blowRatePerFillUnit = WindyLoads.calculateBlowRates(self, spec.lastVehicleSpeed, spec.supportedFillUnits);
	spec.blowRatePerFillUnit = blowRatePerFillUnit; -- store the values
	for fillUnitIndex, fillUnit in pairs(spec.supportedFillUnits) do
		if blowRatePerFillUnit[fillUnitIndex] > 0 or spec.lastBlownVolume[fillUnitIndex] > 0 then
		--we only need to do the calculation if there is something to substract
			local fillLevelPercent = specFillUnit:getFillUnitFillLevelPercentage(fillUnitIndex);
			local fillUnitFillType = specFillUnit:getFillUnitFillType(fillUnitIndex);
			if fillLevelPercent and fillUnitFillType and spec.blowRatePerFillUnit[fillUnitIndex] > 0 then
			--at least we have something that we can work with
				local deltaVol = blowRatePerFillUnit[fillUnitIndex] * fillUnit.fillLevel / 100 * dt / 1000;
				--print(string.format("trying to substract %.4f from fillUnit[%s]. dt = %f; blowRate = %.5f; fillLevel = %f", deltaVol, tostring(fillUnitIndex), dt, blowRatePerFillUnit[fillUnitIndex], fillUnit.fillLevel));
				spec.lastBlownVolume[fillUnitIndex] = spec.lastBlownVolume[fillUnitIndex] + deltaVol;
				spec.timeSinceLastSync[fillUnitIndex] = spec.timeSinceLastSync[fillUnitIndex] + dt;
				if spec.timeSinceLastSync[fillUnitIndex] > WindyLoads.SYNC_TIME_INTERVAL_MS then
					WindyLoads.blowFromFillUnit(self, fillUnitIndex, fillUnit.fillLevel-spec.lastBlownVolume[fillUnitIndex], false);
				end;
				
				--local appliedDelta = specFillUnit:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitIndex, deltaVol, specFillUnit:getFillUnitFillType(fillUnitIndex), ToolType.UNDEFINED, nil);
				--print(string.format("appliedDelta = %.5f; toolTypeSupported = %s", appliedDelta, tostring(specFillUnit:getFillUnitSupportsToolTypeAndFillType(fillUnitIndex, ToolType.UNDEFINED, specFillUnit:getFillUnitFillType(fillUnitIndex)))));
			end;
		end;
	end;
end;

--this function is used to actually perform the reduction of the fill unit caused by blown away loads
function WindyLoads:blowFromFillUnit(fillUnitIndex, newVol, sendNoEvent)
	local spec = self.spec_windyLoads;
	local specFillUnit = self.spec_fillUnit;
	
	--if self.isServer then
		local appliedDelta = specFillUnit:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitIndex, newVol - specFillUnit:getFillUnitFillLevel(fillUnitIndex), specFillUnit:getFillUnitFillType(fillUnitIndex), ToolType.UNDEFINED, nil);
		--print(string.format("appliedDelta = %.5f; toolTypeSupported = %s", appliedDelta, tostring(specFillUnit:getFillUnitSupportsToolTypeAndFillType(fillUnitIndex, ToolType.UNDEFINED, specFillUnit:getFillUnitFillType(fillUnitIndex)))));
		self:updateMass();
		spec.lastBlownVolume[fillUnitIndex] = 0;
		spec.timeSinceLastSync[fillUnitIndex] = 0;
	--end;

	--this is a try to synchronize the fill unit on the client side to the calculations on the server side
	-- if sendNoEvent then
		-- specFillUnit:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitIndex, newVol - specFillUnit:getFillUnitFillLevel(fillUnitIndex), specFillUnit:getFillUnitFillType(fillUnitIndex), ToolType.UNDEFINED, nil);
		-- self:updateMass();
		-- spec.lastBlownVolume[fillUnitIndex] = 0;
		-- spec.timeSinceLastSync[fillUnitIndex] = 0;
	-- else
	if not sendNoEvent then
		WindyLoadsBlowEvent.sendEvent(self, fillUnitIndex, newVol, sendNoEvent);
	end;
end;

function WindyLoads:onUpdate(dt, isActiveForInput, isActiveForInputIngnoreSelection, isSelected)
	local spec = self.spec_windyLoads;
	if not spec.modInitialized or (spec.vehicle ~= nil and spec.vehicleSupported == false) then
		--there's nothing here for us to compute
		return;
	elseif spec.vehicleSupported == nil then
	-- the vehicle should be checked, whether it is applicable for this mod
		spec.vehicleSupported, spec.supportedFillUnits = self:getVehicleIsSupportedByWindyLoads();
		if spec.vehicleSupported == false then
			print(string.format("[WindyLoads]: Vehicle '%s' not supported by WindyLoads mod. It will not be considered any further.", Vehicle.getFullName(self)));
			return;
		end;
	end;
	local specCover = self.spec_cover;
	spec.lastVehicleSpeed = WindyLoads.getVehicleVelocity(self);
	WindyLoads.calculateCurrentRelativeAirSpeed(self);
	--spec.blowRatePerFillUnit = WindyLoads.calculateBlowRates(self, spec.lastVehicleSpeed, spec.supportedFillUnits);
	WindyLoads.affectFillUnits(self, dt, false)	
	if WindyLoads.DEBUG_MODE then
		g_inputBinding:setActionEventActive(spec.actionEventDebugId, true);
		g_inputBinding:setActionEventTextPriority(spec.actionEventDebugId, GS_PRIO_NORMAL);
		g_inputBinding:setActionEventTextVisibility(spec.actionEventDebugId, true);
		g_inputBinding:setActionEventText(spec.actionEventDebugId, g_i18n:getText("input_WINDY_DEBUG"));
	else
		g_inputBinding:setActionEventActive(spec.actionEventDebugId, false);
		g_inputBinding:setActionEventTextVisibility(spec.actionEventDebugId, false);
	end
end;


function WindyLoads:onDraw()
	local spec = self.spec_windyLoads;
	
	if not spec.modInitialized or not WindyLoads.DEBUG_MODE or not spec.vehicleSupported then
		return;
	end;
	--print("windyLoads Draw called");
	if g_currentMission.controlledVehicle ~= nil then
		--only draw debug HUD when inside vehicle
		local vehicleVel;
		local hudLineLevel = WindyLoads.hudData.hudTextPosY;
		local hudText = "Test String";
		--hudText = string.format("vehicle name: %s", g_currentMission.controlledVehicle:getRootVehicle().name);
		hudText = string.format("vehicle: v=%.1f km/h@ X=%.3f; Z=%.3f; Wind: v=%.3f km/h@ X=%.3f; Z=%.3f", 
			spec.lastVehicleSpeed,
			spec.debugInfo.vehicleWorldDirectionX,
			spec.debugInfo.vehicleWorldDirectionZ,
			spec.debugInfo.windVelocity,
			spec.debugInfo.windDirX,
			spec.debugInfo.windDirZ);
			
		setTextColor(0, 1, 0, 1);
		setTextAlignment(RenderText.ALIGN_RIGHT);
		setTextVerticalAlignment(RenderText.VERTICAL_ALIGN_TOP);
		setTextBold(true);
		--renderText(spec.hudData.hudTextPosX, spec.hudData.hudTextPosY, spec.hudData.hudTextSize, "Test Text");
		renderText(WindyLoads.hudData.hudTextPosX, hudLineLevel, WindyLoads.hudData.hudTextSize, hudText);
		
		hudLineLevel = hudLineLevel - WindyLoads.hudData.hudTextSize*1.1;
		hudText = string.format("airSpeed = %.4f", spec.lastAirSpeed);
		renderText(WindyLoads.hudData.hudTextPosX, hudLineLevel, WindyLoads.hudData.hudTextSize, hudText);
		
		for fillUnitIndex, fillUnit in pairs(spec.supportedFillUnits) do
			hudLineLevel = hudLineLevel - WindyLoads.hudData.hudTextSize*1.1;
			hudText = string.format("fillUnit [%s]: fillLevel = %.2f; blowRate = %.4f", tostring(fillUnitIndex), fillUnit.fillLevel, spec.blowRatePerFillUnit[fillUnitIndex]);
			renderText(WindyLoads.hudData.hudTextPosX, hudLineLevel, WindyLoads.hudData.hudTextSize, hudText);
		end;

		
		
		--print("windyLoads HUD Text rendered");
		setTextColor(1,1,1,1);
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextVerticalAlignment(RenderText.VERTICAL_ALIGN_BASELINE);
		setTextBold(false);
	end;
end;


--function to dump tables for debugging purposes
---@param node table The table that is to be printed.
---@param varName string An optional String to represent the name of the printed node. It will be prepended to the output of its contents.
---@param prependMessage string An optional string that is prepended to the out. It will be appended separated with a line break.
function WindyLoads:print_table(node, varName, prependMessage)
    local cache, stack, output = {},{},{}
    local depth = 1
	local varPrintName;
	if varName == nil then
		varPrintName = "";
	else
		varPrintName = varName .. "= ";
	end
    local output_str = varPrintName .. "{\n";
	if prependMessage ~= nil then
		output_str = prependMessage .. "\n" .. output_str;
	end
    while true do
        local size = 0
        for k,v in pairs(node) do
            size = size + 1
        end

        local cur_index = 1
        for k,v in pairs(node) do
            if (cache[node] == nil) or (cur_index >= cache[node]) then

                if (string.find(output_str,"}",output_str:len())) then
                    output_str = output_str .. ",\n"
                elseif not (string.find(output_str,"\n",output_str:len())) then
                    output_str = output_str .. "\n"
                end

                -- This is necessary for working with HUGE tables otherwise we run out of memory using concat on huge strings
                table.insert(output,output_str)
                output_str = ""

                local key
                if (type(k) == "number" or type(k) == "boolean") then
                    key = "["..tostring(k).."]"
                else
                    key = "['"..tostring(k).."']"
                end

                if (type(v) == "number" or type(v) == "boolean") then
                    output_str = output_str .. string.rep('\t',depth) .. key .. " = "..tostring(v)
                elseif (type(v) == "table") then
                    output_str = output_str .. string.rep('\t',depth) .. key .. " = {\n"
                    table.insert(stack,node)
                    table.insert(stack,v)
                    cache[node] = cur_index+1
                    break
                else
                    output_str = output_str .. string.rep('\t',depth) .. key .. " = '"..tostring(v).."'"
                end

                if (cur_index == size) then
                    output_str = output_str .. "\n" .. string.rep('\t',depth-1) .. "}"
                else
                    output_str = output_str .. ","
                end
            else
                -- close the table
                if (cur_index == size) then
                    output_str = output_str .. "\n" .. string.rep('\t',depth-1) .. "}"
                end
            end

            cur_index = cur_index + 1
        end

        if (size == 0) then
            output_str = output_str .. "\n" .. string.rep('\t',depth-1) .. "}"
        end

        if (#stack > 0) then
            node = stack[#stack]
            stack[#stack] = nil
            depth = cache[node] == nil and depth + 1 or depth - 1
        else
            break
        end
    end

    -- This is necessary for working with HUGE tables otherwise we run out of memory using concat on huge strings
    table.insert(output,output_str)
	for k,v in ipairs(output) do
		print(v);
	end
    --output_str = table.concat(output)

    --print(output_str)
end

